﻿//
// file : tinycmp.exe
//
// rev  : 1
// date : 2019/12/08
// auth : ICHIMARU Takeshi(ayakawa.m@gmail.com)
// memo : 初出
//
// 指定したフォルダ同士を比較して
//   フォルダの内容
//   タイムスタンプ
//   ファイルであればbainary
// が同じかどうかをチェックします。
// 同じであれば０を、そうで無ければ１以上を返します。
// なお、ディレクトリのタイムスタンプは復元されないので、チェックしません。
// もしチェックする必要があるならば CHECK_DIR_TIMESTAMP をcompare_dir関数の前で定義してください
//
#include <windows.h>

#include <iostream>
#include <fstream>

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

int check_folder(const char* trg)
{
	struct _stat sts;
	int result = _stat(trg, &sts);

	if (result != 0) {
		// error !
		switch (errno) {
		case ENOENT:
			std::cout << "directory <" << trg << "> not found.\n";
			break;
		case EINVAL:
			std::cout << "<" << trg << "> is invalid value.\n";
			break;
		default:
			std::cout << "expected error in _stat.\n";
		}
		return 0;
	} else {
		// ok. tthen, check directory
		if (_S_IFDIR == (sts.st_mode & _S_IFDIR)) {
			// directoryだったので真
			return 1;
		} else {
			// directoryでは無かった
			std::cout << "<" << trg << "> is not directory.\n";
			return 0;
		}
	}
}

enum CHK_TYPE {
	CHK_FULL,
	CHK_EXIST_ONLY
};

#define BUFSIZE 1000
int compare_file(const char* src, const char* dst, CHK_TYPE chktype) 
{
	struct _stat ss, ds;

	// dstの存在確認
	int result = _stat(dst, &ds);
	if (result != 0) {
		switch (errno) {
		case ENOENT:
			std::cout << "<" << dst << "> not found.\n";
			break;
		default:
			std::cout << "cannot get dstfile info: <" << dst << ">\n";
		}
		return 0;
	}
	if (CHK_EXIST_ONLY == chktype) return 1;  // 存在確認だけの場合はこれで終了

	// 以下、フルの確認
	// dstがファイル？
	if (_S_IFDIR == (ds.st_mode & _S_IFDIR)) {
		std::cout << "we want file, but <" << dst << "> is directory.\n";
		return 0; // dstがディレクトリだったので終了
	}

	// 双方のサイズと最終書き込み日時の比較
	result = _stat(src, &ss);
	if (result != 0) {
		std::cout << "cannot get srcfile info: <" << src << ">\n";
		return 0;
	};
	if (ss.st_size != ds.st_size) {
		std::cout << "different filesize: <" << src << "> / <" << dst << ">\n";
		return 0;
	}
	if (ss.st_mtime != ds.st_mtime) {
		std::cout << "different lwtime: <" << src << "> / <" << dst << ">\n";
		return 0;
	}

	// バイナリの比較
	if (ss.st_size == 0) return 1; // サイズ０なら何もしないで終わり！

	std::streamsize remainsize = ss.st_size,readsize;
	char sbuf[BUFSIZE], dbuf[BUFSIZE];
	std::ifstream s(src, std::ios::in | std::ios::binary), d(dst, std::ios::in | std::ios::binary);

	if (!s) {
		std::cout << "cannot open srcfile: <" << src << ">\n";
		return 0;
	}
	if (!d) {
		std::cout << "cannot open dstfile: <" << dst << ">\n";
		return 0;
	}

	result = 1;
	do {
		readsize = BUFSIZE > remainsize ? remainsize : BUFSIZE;
		remainsize -= readsize;
		std::streamsize ssize = s.read(&sbuf[0], readsize).gcount();
		std::streamsize dsize = d.read(&dbuf[0], readsize).gcount();
		if (ssize != dsize || ssize != readsize) {
			// 読み込み失敗してるよ！
			std::cout << "fail to read file: read=" << readsize << "/src=" << ssize << "/dst=" << dsize << "\n";
			result = 0;
			break;
		}
		// 比較
		if (memcmp(&sbuf[0], &dbuf[0], (size_t)readsize)) {
			// データが違う！
			std::cout << "different data: <" << src << "> / <" << dst << ">\n";
			result = 0;
			break;
		}
	} while (remainsize>0);

	return result;
}

//
// #define CHECK_DIR_TIMESTAMP
//
int compare_dir(const char* src, const char* dst, CHK_TYPE chktype)
{
	HANDLE hFind;
	WIN32_FIND_DATA wfd;

	std::string src_path = src, dst_path = dst;
	std::string search_path = src_path + "\\*";

	hFind = FindFirstFile(search_path.c_str(), &wfd);

	if (INVALID_HANDLE_VALUE == hFind) {
		// 見つからなかった？？
		std::cout << "cannot search <" << src_path << ">\n";
		return 0;
	}

	int all_result = 1;
	do {
		if (FILE_ATTRIBUTE_DIRECTORY==(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
			// direcotryの場合
			std::string filename = wfd.cFileName;
			if (filename != "." && filename != "..") {
				// . と .. 以外の場合、dstに有るかを確認
				std::string dst_search_path = dst;
				dst_search_path += "\\"; dst_search_path += filename;
				struct _stat sts;
				int result = _stat(dst_search_path.c_str(),&sts);
				if (result != 0) {
					// 情報の取得が出来ない
					switch (errno) {
						case ENOENT:
							std::cout << "dst:<" << dst_search_path << "> not found.\n";
							break;
						default:
							std::cout << "cannot get dstfile stat : <" << dst_search_path << ">\n";
							break;
					}
					all_result = 0;
					// 次のファイル/ディレクトリのチェックへ
				} else {
					// 存在確認のみなので、次へ
					if (CHK_EXIST_ONLY == chktype) {
						// 何か有れば良いので、何もしないで次へ
						;
					} else {
						// 以下、フルチェック
						// directory?
						if (_S_IFDIR == (sts.st_mode & _S_IFDIR)) {
							// direcotryだった
							std::string src_search_path = src;
							src_search_path += "\\"; src_search_path += filename;
#ifdef CHECK_DIR_TIMESTAMP
							// フォルダのタイムスタンプは再現されないのでチェックしません
							struct _stat s_sts;
							result = _stat(src_search_path.c_str(), &s_sts);
							// compare last write time
							if (sts.st_mtime != s_sts.st_mtime) {
								// 等しくなかった！
								all_result = 0;
								std::cout << "different lwtime: <" << src_search_path << "> / <" << dst_search_path << ">\n";
								// フォルダなので、サブフォルダ以下のチェックは続行する
							}
#endif
							// サブフォルダのチェックを行う
							if (0 == compare_dir(src_search_path.c_str(), dst_search_path.c_str(), chktype)) all_result = 0;
						}
						else {
							// directoryでは無い
							std::cout << "src: <" << src << "\\" << filename << "> is directory,"
								<< " but dst: <" << dst_search_path << "> is not directory\n";
							all_result = 0;
							// エラーは確定。次へ続行
						}
					}
				}
			}
		}
		else {
			// fileの場合
			std::string src_file = src, dst_file = dst;
			src_file += "\\"; src_file += wfd.cFileName;
			dst_file += "\\"; dst_file += wfd.cFileName;
			if (0==compare_file(src_file.c_str(), dst_file.c_str(), chktype)) all_result = 0;
		}
	} while (FindNextFile(hFind, &wfd));

	FindClose(hFind);

	return all_result;
}

int main(int argc, char *argv[])
{
	// check arg - usage: tinycmp src dst
	if (3 != argc) {
		std::cout << "usage: tinycmp srcdir dstdir\n";
		return 1;
	}

	// check src is folder ?
	if (!check_folder(argv[1])) return 2;

	// check dst is folder ?
	if (!check_folder(argv[2])) return 3;

	int result = 0;
	// compare src -> dst !! srcにあってdstに無い/タイムスタンプが違う/サイズが違う/中身が違う/ファイルの種類が違う、がエラーになる
	int sts = compare_dir(argv[1], argv[2], CHK_FULL);
	if (!sts) result = 4;

	// compare dst -> src !! dstにあってsrcに無いケースのみがエラーになる
	sts = compare_dir(argv[2], argv[1], CHK_EXIST_ONLY);
	if (!sts) {
		if (4 == result) {
			result = 6;
		} else {
			result = 5;
		}
	}

	// good !!
 	return result;
}

// プログラムの実行: Ctrl + F5 または [デバッグ] > [デバッグなしで開始] メニュー
// プログラムのデバッグ: F5 または [デバッグ] > [デバッグの開始] メニュー

// 作業を開始するためのヒント: 
//    1. ソリューション エクスプローラー ウィンドウを使用してファイルを追加/管理します 
//   2. チーム エクスプローラー ウィンドウを使用してソース管理に接続します
//   3. 出力ウィンドウを使用して、ビルド出力とその他のメッセージを表示します
//   4. エラー一覧ウィンドウを使用してエラーを表示します
//   5. [プロジェクト] > [新しい項目の追加] と移動して新しいコード ファイルを作成するか、[プロジェクト] > [既存の項目の追加] と移動して既存のコード ファイルをプロジェクトに追加します
//   6. 後ほどこのプロジェクトを再び開く場合、[ファイル] > [開く] > [プロジェクト] と移動して .sln ファイルを選択します
